package grm;

import grm.ExecutionManagerImpl.AppInfo;
import interfaces.Grm;
import interfaces.Lrm;
import interfaces.LrmHelper;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.atomic.AtomicInteger;

import org.omg.CORBA.ORB;

import subtypes.ApplicationType;
import subtypes.AsctRequestId;
import types.BspInfo;
import types.CommonExecutionSpecs;
import types.DistinctExecutionSpecs;

/** A class is created for every restart */
public class BspRestartCoordinator {
    
    private ORB orb;
    private Grm grm;
    private int nProcesses;
    
    private BspFindGlobalCkpNumber findCkpNumber;
    private CyclicBarrier barrier;
    private String bspProcessZeroIor;    
        
    //---------------------------------------------------------------------------------------
    
    public BspRestartCoordinator(ORB orb, Grm grm, int nProcesses) {
        bspProcessZeroIor = null;                
        barrier = new CyclicBarrier(nProcesses);
        this.orb = orb; this.grm = grm;
        this.nProcesses = nProcesses;
    }
    
    //---------------------------------------------------------------------------------------
    public void restartBspApplication(AsctRequestId asctRequestId, AppInfo appInfo, boolean restartAll) {        

        if (restartAll == true) {
            findCkpNumber = new BspFindGlobalCkpNumber(nProcesses);
            appInfo.nRestarts.set(0);
            appInfo.nResponses.set(0);
            int nWaiting = barrier.getNumberWaiting();            
            barrier.reset();
            if (nWaiting > 0)
                barrier = new CyclicBarrier(nProcesses);
        }
        else {
            if (appInfo.nResponses.get() < appInfo.distinctSpecs.length) {  
              /** Wait until all LRM respond to the restart method */
              appInfo.waitingResponses.lock();
              try {appInfo.finishedResponsesCond.await();}
              catch (Exception e) {e.printStackTrace();}
              finally{appInfo.waitingResponses.unlock();}
            }
            /** Descrements nResponses, so that other reschedules will wait */
            appInfo.nResponses.decrementAndGet();
        }

        if (ExecManDebug.Debug)
            System.err.println("<<<-->>> BspRestartCoordinator.restartBspApplication-->asctRequestId: " + asctRequestId.appMainRequestId + "|" + asctRequestId.appNodeRequestId + " nRestart: {" + appInfo.nRestarts + "|" + appInfo.nResponses + "}");

        /** Reschedules the killed process for execution with the GRM */        
        Integer failedNodeId = Integer.valueOf(asctRequestId.appNodeRequestId);                                                                   
        new BspRescheduleExecutionThread(orb, grm, findCkpNumber, appInfo, failedNodeId).start();
        appInfo.nodeAppIds.remove(failedNodeId);
        
        /** Restart the remaining processes in their currently executing nodes */
        if (restartAll == true) {        
            Iterator<Integer> it = appInfo.nodeAppIds.keySet().iterator();                                   
            while (it.hasNext())
                new BspRestartExecutionThread(orb, findCkpNumber, appInfo, it.next()).start();        
        }
               
    }
    
    //-------------------------------------------------------------------------------------
    
    public BspInfo registerBspNode(AsctRequestId asctRequestId, String bspProxyIor){                
        
        int nodeId = Integer.parseInt(asctRequestId.appNodeRequestId);        
        if(nodeId == 0) bspProcessZeroIor = bspProxyIor;
        
        /** Synchronization point */
        if (ExecManDebug.Debug) 
            System.err.println("RegisterBspNode: waiting for BSP processes, nodeId: " + nodeId);
        try {barrier.await();}
        catch(BrokenBarrierException broken) {
            System.err.println("RegisterBspNode: Barrier was broken. @#$%@#$%@#$%@#$%@#$%@#$%@#$%");
        }
        catch(Exception e) {e.printStackTrace();};
        if (nodeId == 0)
            System.err.println("RegisterBspNode: returning Process Zero IOR");               
        
        if (nodeId == 0) return new BspInfo(true, bspProcessZeroIor);
        else return new BspInfo(false, bspProcessZeroIor);
        
    }
}

/** ---------------------------------------------------------------------------
 *  BspFindGlobalCkpNumber Class
 *  ---------------------------------------------------------------------------
 * */

class CkpNumberNotAvailable extends Exception {
    private static final long serialVersionUID = -116761624795937772L;        
}

class BspFindGlobalCkpNumber { 
    
    Vector<Integer> ckpNumbers;
    int nTotal;
    int ckpNumber_;
    boolean ckpNumberAvailable = false;
    
    public BspFindGlobalCkpNumber(int nProcesses) {
        ckpNumbers = new Vector<Integer>(nProcesses);
        nTotal = nProcesses;
    }
    
    public int getCkpNumber () throws CkpNumberNotAvailable {
        if (ckpNumberAvailable == false) throw new CkpNumberNotAvailable();
        return ckpNumber_;
    }
    
    /** Adds a checkpoint number from a execution. 
     *  It automatically synchronizes the threads so that they are realeased only
     *  after a checkpoint mumber for the recovery has been determined */
    public synchronized int addCkpNumber(int ckpNumber) {
        ckpNumbers.add(new Integer(ckpNumber));
        //System.out.println("ckpNumber before:" + ckpNumber);
        int nCkp = ckpNumbers.size();
        try {
            if (nCkp < nTotal)
                wait();
            else if (nCkp == nTotal)
                findCkpNumber();
            notifyAll();
        } catch (Exception e) {
            e.printStackTrace();
        }
        //System.out.println("ckpNumber after:" + ckpNumber_);
        return ckpNumber_;
    }
    
    private void findCkpNumber () {
        ckpNumber_ = Integer.MAX_VALUE;        
        for (int i=0; i < ckpNumbers.size(); i++) {
            if ((ckpNumbers.get(i)).intValue() < ckpNumber_ )
                ckpNumber_ = ckpNumbers.get(i).intValue();                
        }        
        ckpNumberAvailable = true;
    }
}


/** ---------------------------------------------------------------------------
 *  BspRescheduleExecutionThread Class
 *  ---------------------------------------------------------------------------
 * */

class BspRescheduleExecutionThread extends Thread {
    
    private Grm grm;
    private DistinctExecutionSpecs distinctSpecs;
    private CommonExecutionSpecs commonSpecs;
    private boolean forceDifferentMachines;
    
    private AppInfo appInfo;
    private Lrm lrm;
    private String lrmIOR;
    private String appId;
    private BspFindGlobalCkpNumber findCkpNumber;
    private AtomicInteger nResponses;
    
    BspRescheduleExecutionThread (ORB orb, Grm grm, BspFindGlobalCkpNumber findCkpNumber,
                                  AppInfo appInfo, int nodeIndex) {

        this.appInfo = appInfo;
        this.commonSpecs = appInfo.commonSpecs;
        this.distinctSpecs = appInfo.distinctSpecs[nodeIndex];
        this.forceDifferentMachines = appInfo.forceDifferentMachines;
        this.nResponses = appInfo.nResponses;
        String[] nodeInfo = appInfo.nodeAppIds.get(nodeIndex);
        this.lrmIOR = nodeInfo[0];
        this.appId  = nodeInfo[1];
        
        this.grm = grm;        
        this.lrm = LrmHelper.narrow( orb.string_to_object( lrmIOR ));
        this.findCkpNumber = findCkpNumber;

    }
    
    public void run() {
        
        int ckpNumber;
        try { ckpNumber = findCkpNumber.getCkpNumber(); } 
        catch (CkpNumberNotAvailable e) {
            int lastCkp = lrm.getLastCkpNumber(appId);        
            ckpNumber = findCkpNumber.addCkpNumber(lastCkp); // Automatically Synchronized
        } 
             
        if (ckpNumber >=0) {            
            String ckpName = " -r" + ckpNumber + " -d../" + appId;
            distinctSpecs.applicationArgs = ckpName;
        }
                
        /** Used in experiments to force reinitialization on the same machine */
        String appConstraints = commonSpecs.applicationConstraints;
        commonSpecs.applicationConstraints.trim(); 
        if(commonSpecs.applicationConstraints.length()>0)
            commonSpecs.applicationConstraints = commonSpecs.applicationConstraints + " and lrmIOR == '" + lrmIOR + "' ";
        else
            commonSpecs.applicationConstraints = commonSpecs.applicationConstraints + " lrmIOR == '" + lrmIOR + "' ";

        /** Performs the remote execution request */
        grm.remoteExecutionRequest(commonSpecs, new DistinctExecutionSpecs[] {distinctSpecs}, forceDifferentMachines);        
        
        /** Return to previous constraint */
        commonSpecs.applicationConstraints = appConstraints;

        int nResp = nResponses.incrementAndGet();
        if (nResp == appInfo.distinctSpecs.length) {
            appInfo.waitingResponses.lock();
            try {appInfo.finishedResponsesCond.signal();}
            finally{appInfo.waitingResponses.unlock();}
        }
        
        if (ExecManDebug.Debug)
            System.err.println(">>>>> RescheduleExecutionThread --> " + appId + " ckpNumber=" + ckpNumber + " nResponses=" + nResponses);
    }
    
}

/** ---------------------------------------------------------------------------
 *  BspRestartExecutionThread Class
 *  ---------------------------------------------------------------------------
 * */

class BspRestartExecutionThread extends Thread {
    
    private AppInfo appInfo;
    private Lrm lrm;
    private String appId;
    private BspFindGlobalCkpNumber findCkpNumber;
    private DistinctExecutionSpecs distinctSpecs;
    private AtomicInteger nRestarts;
    private AtomicInteger nResponses;    
    
    BspRestartExecutionThread(ORB orb, BspFindGlobalCkpNumber findCkpNumber,
                              AppInfo appInfo, int nodeIndex) {

        this.appInfo = appInfo;
        this.distinctSpecs = appInfo.distinctSpecs[nodeIndex];
        this.nRestarts  = appInfo.nRestarts;
        this.nResponses = appInfo.nResponses;
        String[] nodeInfo = appInfo.nodeAppIds.get(nodeIndex);
        this.appId  = nodeInfo[1];
               
        this.lrm = LrmHelper.narrow( orb.string_to_object( nodeInfo[0] ));
        this.findCkpNumber = findCkpNumber;        
    }
    
    public void run() {
        int lastCkp   = lrm.getLastCkpNumber(appId);        
        int ckpNumber = findCkpNumber.addCkpNumber(lastCkp); // Automatically Synchronized
        
        if (ckpNumber >=0) {
            String ckpName = " -r" + ckpNumber + " -d../" + appId;
            distinctSpecs.applicationArgs = ckpName;
        }
        
        //String[] nodeAppIdInfo = appInfo.nodeAppIds.remove(nodeAppId);
        //assert(nodeAppIdInfo != null);
        int restartFlag = lrm.restartExecution(appId, distinctSpecs.applicationArgs);
        /** Restarting was succesful */
        //if (restartFlag==0) appInfo.nodeAppIds.put(nodeAppId, nodeAppIdInfo);
        if (restartFlag==0) nRestarts.incrementAndGet();

        int nResp = nResponses.incrementAndGet();
        if (nResp == appInfo.distinctSpecs.length) {
            appInfo.waitingResponses.lock();
            try {appInfo.finishedResponsesCond.signal();}
            finally{appInfo.waitingResponses.unlock();}
        }
        
        if (ExecManDebug.Debug)
            System.err.println(">>>>> RestartExecutionThread --> " + appId + "|" + restartFlag + " ckpNumber=" + ckpNumber  + " nResponses=" + nResponses + " nRestarts=" + nRestarts);
    }
}

//-----------------------------------------------------------------------------------------
